/*
    pmacct (Promiscuous mode IP Accounting package)
    pmacct is Copyright (C) 2004 by Paolo Lucente
*/

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#define __NET_AGGR_C

/* includes */
#include "pmacct.h"
#include "nfacctd.h"
#include "pmacct-data.h"
#include "plugin_hooks.h"
#include "net_aggr.h"
#include "util.h"

void load_networks(char *filename, struct networks_table *nt, struct networks_cache *nc)
{
  FILE *file;
  struct networks_table tmp, *tmpt = &tmp; 
  struct networks_table_metadata *mdt;
  char buf[SRVBUFLEN], *bufptr, *delim, *as, *net, *mask;
  int rows, index, eff_rows = 0;

  memset(nt, 0, sizeof(struct networks_table));
  memset(nc, 0, sizeof(struct networks_cache));
  memset(&dummy_entry, 0, sizeof(struct networks_cache));

  if (filename) {
    if ((file = fopen(filename,"r")) == NULL) {
      if ((config.acct_type == ACCT_NF) && (config.what_to_count & (COUNT_SRC_AS|COUNT_DST_AS)) &&
	  (!(config.what_to_count & (COUNT_SRC_NET|COUNT_DST_NET))) && (config.nfacctd_as == NF_AS_KEEP))
	return;

      Log(LOG_ERR, "ERROR: network file '%s' not found\n", filename);
      exit(1); 
    }
    else {
      rows = 0;
      /* 1st step: count rows for table allocation */
      while (!feof(file)) {
        if (fgets(buf, SRVBUFLEN, file)) rows++; 
      }
      /* 2nd step: loading data into a temporary table */
      freopen(filename, "r", file);
      nt->table = malloc(rows*sizeof(struct networks_table_entry)); 
      tmpt->table = malloc(rows*sizeof(struct networks_table_entry));
      memset(nt->table, 0, rows*sizeof(struct networks_table_entry));
      memset(tmpt->table, 0, rows*sizeof(struct networks_table_entry));
      rows = 1;
      while (!feof(file)) {
	bufptr = buf;
	memset(buf, 0, SRVBUFLEN);
        if (fgets(buf, SRVBUFLEN, file)) { 
	  if (delim = strchr(buf, ',')) {
	    as = buf;
	    *delim = '\0';
	    bufptr = delim+1;
	    tmpt->table[eff_rows].as = atoi(as);
	  }
	  if (!sanitize_buf_net(bufptr, rows)) {
	    delim = strchr(bufptr, '/');
	    *delim = '\0';
	    net = bufptr;
	    mask = delim+1;
	    if (!strchr(mask, '.')) {
	      index = atoi(mask); 
	      if ((index >= 0) && (index <= 32)) mask = cidr_table[index].value; 
	      else {
		mask = cidr_table[0].value; /* we will point 'mask' to an invalid value */
		Log(LOG_ERR, "ERROR: invalid network mask '/%d'\n", index);
	      }
	    }
	    inet_aton(net, (struct in_addr *) &tmpt->table[eff_rows].net);
	    inet_aton(mask, (struct in_addr *) &tmpt->table[eff_rows].mask);

	    tmpt->table[eff_rows].net = ntohl(tmpt->table[eff_rows].net);
	    tmpt->table[eff_rows].mask = ntohl(tmpt->table[eff_rows].mask);

	    /* we will insert in the networks table only valid values; 0.0.0.0/0
	       is not and we silently discard it */ 
	    if ((!tmpt->table[eff_rows].net) || (!tmpt->table[eff_rows].mask));
	    else eff_rows++;
	  } 
	}
	rows++;
      }
      fclose(file);

      /* 3rd step: sorting table */
      merge_sort(tmpt->table, 0, eff_rows);
      tmpt->num = eff_rows;

      /* 4th step: collecting informations in the sorted table;
         we wish to handle networks-in-networks hierarchically */

      /* 4a building hierarchies */
      mdt = malloc(tmpt->num*sizeof(struct networks_table_metadata));
      memset(mdt, 0, tmpt->num*sizeof(struct networks_table_metadata));

      for (index = 0; index < (tmpt->num-1); index++) {
	u_int32_t net;
	int x;

	for (x = index+1; x < tmpt->num; x++) {
	  net = tmpt->table[x].net;
	  net &= tmpt->table[index].mask;
	  if (net == tmpt->table[index].net) {
	    mdt[x].level++;
	    mdt[index].childs++; 
	  }
	  else break;
	}
      } 

      /* 4b retrieving root entries number */
      for (index = 0, eff_rows = 0; index < tmpt->num; index++) {
        if (mdt[index].level == 0) eff_rows++;
      }

      nt->num = eff_rows;
      /* 4c adjusting child counters: each parent has to know
         only the number of its directly attached childs and
	 not the whole hierarchy */ 
      for (index = 0; index < tmpt->num; index++) {
        int x, eff_childs = 0;

        for (x = index+1; x < tmpt->num; x++) {
	  if (mdt[index].level == mdt[x].level) break;
	  else if (mdt[index].level == (mdt[x].level-1)) eff_childs++; 
	}
	mdt[index].childs = eff_childs;
      }

      /* 5th step: building final networks table */
      for (index = 0; index < tmpt->num; index++) {
	int current, next, prev[128];

        if (!index) {
	  current = 0; next = eff_rows;
	  memset(&prev, 0, 32);
	  memcpy(&nt->table[current], &tmpt->table[index], sizeof(struct networks_table_entry));
        }
	else {
	  if (mdt[index].level == mdt[index-1].level) current++; /* do nothing: we have only to copy our element */ 
	  else if (mdt[index].level > mdt[index-1].level) { /* we encountered a child */ 
	    nt->table[current].childs_table.table = &nt->table[next];
	    nt->table[current].childs_table.num = mdt[index-1].childs;
	    prev[mdt[index-1].level] = current;
	    current = next;
	    next += mdt[index-1].childs;
	  }
	  else { /* going back to parent level */
	    current = prev[mdt[index].level];
	    current++;
	  }
	  memcpy(&nt->table[current], &tmpt->table[index], sizeof(struct networks_table_entry));
        }
      }

      if (config.debug) { 
        index = 0;
        while (index < tmpt->num) {
	  Log(LOG_DEBUG, "DEBUG: (networks table) net: %x, mask: %x\n", nt->table[index].net, nt->table[index].mask); 
	  // Log(LOG_DEBUG, "DEBUG: (networks table) net: %x, mask: %x, level: %u, childs: %u\n",
	  //   tmpt->table[index].net, tmpt->table[index].mask, mdt[index].level, mdt[index].childs); 
	  index++;
	}
      }

      /* create networks cache */
      if (!config.networks_cache_entries) nc->num = NETWORKS_CACHE_ENTRIES;
      else nc->num = config.networks_cache_entries;
      nc->cache = (struct networks_cache_entry *) malloc(nc->num*sizeof(struct networks_cache_entry));
      memset(nc->cache, 0, nc->num*sizeof(struct networks_cache_entry));
    }
  }
}

/* sort the (sub)array v from start to end */
void merge_sort(struct networks_table_entry *table, int start, int end)
{
  int middle;

  /* no elements to sort */
  if ((start == end) || (start == end-1)) return;

  /* find the middle of the array, splitting it into two subarrays */
  middle = (start+end)/2;

  /* sort the subarray from start..middle */
  merge_sort(table, start, middle);

  /* sort the subarray from middle..end */
  merge_sort(table, middle, end);

  /* merge the two sorted halves */
  merge(table, start, middle, end);
}

/* 
   merge the subarray v[start..middle] with v[middle..end], placing the
   result back into v.
*/
void merge(struct networks_table_entry *table, int start, int middle, int end)
{
  struct networks_table_entry *v1, *v2;
  int  v1_n, v2_n, v1_index, v2_index, i, s = sizeof(struct networks_table_entry);

  v1_n = middle-start;
  v2_n = end-middle;

  v1 = malloc(v1_n*s);
  v2 = malloc(v2_n*s);

  if ((!v1) || (!v2)) Log(LOG_ERR, "ERROR: Memory sold out when allocating 'networks table'\n"); 

  for (i=0; i<v1_n; i++) memcpy(&v1[i], &table[start+i], s);
  for (i=0; i<v2_n; i++) memcpy(&v2[i], &table[middle+i], s);

  v1_index = 0;
  v2_index = 0;

  /* as we pick elements from one or the other to place back into the table */
  for (i=0; (v1_index < v1_n) && (v2_index < v2_n); i++) {
    /* current v1 element less than current v2 element? */
    if (v1[v1_index].net < v2[v2_index].net) memcpy(&table[start+i], &v1[v1_index++], s);
    else if (v1[v1_index].net == v2[v2_index].net) {
      if (v1[v1_index].mask <= v2[v2_index].mask) memcpy(&table[start+i], &v1[v1_index++], s);
      else memcpy(&table[start+i], &v2[v2_index++], s);
    }
    else memcpy(&table[start+i], &v2[v2_index++], s); 
  }

  /* clean up; either v1 or v2 may have stuff left in it */
  for (; v1_index < v1_n; i++) memcpy(&table[start+i], &v1[v1_index++], s);
  for (; v2_index < v2_n; i++) memcpy(&table[start+i], &v2[v2_index++], s);

  free(v1);
  free(v2);
}

struct networks_table_entry *binsearch(struct networks_table *nt, struct networks_cache *nc, struct in_addr *a)
{
  int low = 0, mid, high = nt->num-1;
  u_int32_t net, addrh = ntohl(a->s_addr), addr = a->s_addr;
  struct networks_table_entry *ret;

  ret = networks_cache_search(nc, &addr); 
  if (ret) return ret;

  while (low <= high) {
    mid = (low+high)/2;
    net = addrh;
    net &= nt->table[mid].mask;

    if (net < nt->table[mid].net) {
      high = mid-1;
      continue;
    }
    else if (net > nt->table[mid].net) {
      low = mid+1;
      continue;
    }

    /* It's assumed we've found our element */
    if (nt->table[mid].childs_table.table) {
      ret = binsearch(&nt->table[mid].childs_table, nc, a);
      if (!ret) {
	ret = &nt->table[mid];
        networks_cache_insert(nc, &addr, &nt->table[mid]);
      }
    }
    else {
      ret = &nt->table[mid];
      networks_cache_insert(nc, &addr, &nt->table[mid]);
    }
    return ret;
  }

  networks_cache_insert(nc, &addr, &dummy_entry);
  return NULL;
}

void networks_cache_insert(struct networks_cache *nc, u_int32_t *key, struct networks_table_entry *result)
{
  struct networks_cache_entry *ptr;

  ptr = &nc->cache[*key % nc->num];
  ptr->key = *key;
  ptr->result = result;
}

struct networks_table_entry *networks_cache_search(struct networks_cache *nc, u_int32_t *key)
{
  struct networks_cache_entry *ptr;

  ptr = &nc->cache[*key % nc->num];
  if (ptr->key == *key) return ptr->result;
  else return NULL;
}

void set_net_funcs(struct networks_table *nt)
{
  u_int8_t count = 0;

  memset(&net_funcs, 0, sizeof(net_funcs));

  if (!nt->num) return;

  if (config.what_to_count & (COUNT_SRC_HOST|COUNT_SUM_HOST)) {
    net_funcs[count] = search_src_host;
    count++;
  }

  if (config.what_to_count & (COUNT_SRC_NET|COUNT_SUM_NET)) {
    net_funcs[count] = search_src_net;
    count++;
  } 

  if (config.what_to_count & (COUNT_SRC_AS|COUNT_SUM_AS)) {
    if ((config.acct_type == ACCT_NF) && (config.nfacctd_as == NF_AS_KEEP));
    else {
      net_funcs[count] = search_src_as;
      count++;
    }
  }

  if (config.what_to_count & (COUNT_DST_HOST|COUNT_SUM_HOST)) {
    net_funcs[count] = search_dst_host;
    count++;
  }

  if (config.what_to_count & (COUNT_DST_NET|COUNT_SUM_NET)) {
    net_funcs[count] = search_dst_net;
    count++;
  }

  if (config.what_to_count & (COUNT_DST_AS|COUNT_SUM_AS)) {
    if ((config.acct_type == ACCT_NF) && (config.nfacctd_as == NF_AS_KEEP));
    else {
      net_funcs[count] = search_dst_as;
      count++;
    }
  }
}

void search_src_host(struct networks_table *nt, struct networks_cache *nc, struct pkt_primitives *p)
{
  struct networks_table_entry *res;

  res = binsearch(nt, nc, &p->src_ip);
  if (!res) p->src_ip.s_addr = 0;
  else if (!res->net) p->src_ip.s_addr = 0; /* it may have been cached */
}

void search_dst_host(struct networks_table *nt, struct networks_cache *nc, struct pkt_primitives *p)
{
  struct networks_table_entry *res;

  res = binsearch(nt, nc, &p->dst_ip);
  if (!res) p->dst_ip.s_addr = 0;
  else if (!res->net) p->dst_ip.s_addr = 0; /* it may have been cached */
}

void search_src_net(struct networks_table *nt, struct networks_cache *nc, struct pkt_primitives *p)
{
  struct networks_table_entry *res;

  res = binsearch(nt, nc, &p->src_ip);
  if (!res) p->src_ip.s_addr = 0;
  else p->src_ip.s_addr = htonl(res->net);
}

void search_dst_net(struct networks_table *nt, struct networks_cache *nc, struct pkt_primitives *p)
{
  struct networks_table_entry *res;
  
  res = binsearch(nt, nc, &p->dst_ip);
  if (!res) p->dst_ip.s_addr = 0;
  else p->dst_ip.s_addr = htonl(res->net);
}

void search_src_as(struct networks_table *nt, struct networks_cache *nc, struct pkt_primitives *p)
{
  struct networks_table_entry *res;
  
  res = binsearch(nt, nc, &p->src_ip);
  if (!res) p->src_ip.s_addr = 0;
  else p->src_ip.s_addr = htonl(res->as);
}

void search_dst_as(struct networks_table *nt, struct networks_cache *nc, struct pkt_primitives *p)
{
  struct networks_table_entry *res;
  
  res = binsearch(nt, nc, &p->dst_ip);
  if (!res) p->dst_ip.s_addr = 0;
  else p->dst_ip.s_addr = htonl(res->as);
}
