/*  
    pmacct (Promiscuous mode IP Accounting package)
    pmacct is Copyright (C) 2004 by Paolo Lucente
*/

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#define __PMACCT_CLIENT_C

/* include */
#include "pmacct.h"
#include "pmacct-data.h"
#include "imt_plugin.h"

#define CHECK_Q_TYPE(x) x & WANT_RESET ? x ^ WANT_RESET : x

/* prototypes */
int Recv(int, unsigned char **);
int sanitize_buf(char *);
void trim_all_spaces(char *);
char *extract_token(char **, int);

/* functions */
void usage_client(char *prog)
{
  printf("%s\n", PMACCT_USAGE_HEADER);
  printf("Usage: %s [query]\n\n", prog);
  printf("Queries:\n");
  printf("  -s\tcollect full table statistics\n"); 
  printf("  -N\t[matching data[: ... ]] \n\tmatch exactly actual data; print counter only\n");
  printf("  -n\t[bytes|packets] \n\tselect counter to output (applies to -N)\n");
  printf("  -M\t[matching data[: ... ]] \n\tmatch either partially or exactly actual data; print table\n");
  printf("  -c\t[src_mac|dst_mac|vlan|src_host|dst_host|src_port|dst_port|proto|src_as|dst_as| \n\t |sum_host|sum_net|sum_as|sum_port] \n\ttraffic aggregation (required by -N and -M)\n");
  printf("  -a\tdisplay all fields\n");
  printf("  -e\tclear statistics\n");
  printf("  -r\treset counters for the specified entry \n\t(applies to either -N or -M)\n");
  printf("  -t\tcheck table status\n");
  printf("  -p\t[file] \n\tsocket for client-server communication (DEFAULT: /tmp/collect.pipe)\n");
  printf("\n");
  printf("  See EXAMPLES file in the distribution for examples\n");
  printf("\n");
  printf("For suggestions, critics, bugs, contact me: %s.\n", MANTAINER);
}

int sanitize_buf(char *buf)
{
  int x = 0, valid_char = 0;

  trim_all_spaces(buf);
  while (x < strlen(buf)) {
    if (!isspace(buf[x])) valid_char++;
    x++;
  }
  if (!valid_char) return TRUE;
  if (buf[0] == '!') return TRUE;

  return FALSE;
}

void trim_all_spaces(char *buf)
{
  char *ptr;
  int i = 0, len;

  ptr = buf;
  len = strlen(buf);

  /* trimming all spaces */
  while (i <= len) {
    if (isspace(ptr[i])) {
      strcpy(&buf[i], &ptr[i+1]);
      len--;
    }
    else i++;
  }
}

void print_ex_options_error()
{
  printf("ERROR: '-s', '-t', '-N', '-e' and '-M' options are each mutually exclusive.\n\n");
  exit(1);
}

void write_stats_header(u_int32_t what_to_count, u_int8_t have_wtc)
{
  if (!have_wtc) {
    printf("ID     ");
    printf("SRC MAC            ");
    printf("DST MAC            ");
    printf("VLAN   ");
    printf("SRC IP           ");
    printf("DST IP           ");
    printf("SRC PORT  ");
    printf("DST PORT  ");
    printf("PROTOCOL    ");
    // printf("TOS    ");
    printf("PACKETS     ");
    printf("BYTES\n");
  }
  else {
    if (what_to_count & COUNT_ID) printf("ID     ");
    if (what_to_count & COUNT_SRC_MAC) printf("SRC MAC            "); 
    if (what_to_count & COUNT_DST_MAC) printf("DST MAC            "); 
    if (what_to_count & COUNT_VLAN) printf("VLAN   ");
    if (what_to_count & (COUNT_SRC_HOST|COUNT_SRC_NET|COUNT_SRC_AS)) printf("SRC IP           ");
    if (what_to_count & (COUNT_SUM_HOST|COUNT_SUM_NET|COUNT_SUM_AS)) printf("SRC IP           ");
    if (what_to_count & (COUNT_DST_HOST|COUNT_DST_NET|COUNT_DST_AS)) printf("DST IP           ");
    if (what_to_count & (COUNT_SRC_PORT|COUNT_SUM_PORT)) printf("SRC PORT  ");
    if (what_to_count & COUNT_DST_PORT) printf("DST PORT  "); 
    if (what_to_count & COUNT_IP_PROTO) printf("PROTOCOL    ");
    // if (what_to_count & COUNT_IP_TOS) printf("TOS    ");
    printf("PACKETS     ");
    printf("BYTES\n");
  }
}

void write_status_header()
{
  printf("* = element\n\n"); 
  printf("BUCKET\tCHAIN STATUS\n");
}

int build_query_client(char *path_ptr)
{
  struct sockaddr_un cAddr;
  int sd, rc, cLen;

  sd = socket(AF_UNIX, SOCK_STREAM, 0);
  if (sd < 0) {
    printf("ERROR: Unable to open socket.\n");
    exit(1);
  }

  cAddr.sun_family = AF_UNIX;
  strcpy(cAddr.sun_path, path_ptr);
  cLen = sizeof(cAddr);

  rc = connect(sd, (struct sockaddr *) &cAddr, cLen);
  if (rc < 0) {
    if (errno == ECONNREFUSED) {
      printf("INFO: Connection refused while trying to connect to '%s'\n\n", path_ptr);
      exit(0);
    }
    else {
      printf("ERROR: Unable to connect to '%s'\n\n", path_ptr);
      exit(1);
    }
  }

  return sd;
}

int main(int argc,char **argv)
{
  int clibufsz = (MAX_QUERIES*sizeof(struct pkt_primitives))+sizeof(struct query_header)+2;
  struct acc *acc_elem;
  struct bucket_desc *bd;
  struct query_header q; 
  struct pkt_primitives empty_addr, addr;
  char clibuf[clibufsz], *bufptr;
  unsigned char *largebuf, *elem;
  char *src_ip;
  char path[128], file[128], password[9];
  int sd, num, buflen, unpacked, printed;
  int counter=0;

  /* mrtg stuff */
  char match_string[LARGEBUFLEN], *match_string_token, *match_string_ptr;
  char count[128], *count_token[N_PRIMITIVES], *count_ptr;
  int count_index = 0, match_string_index = 0, index = 0;
  
  /* getopt() stuff */
  extern char *optarg;
  extern int optind, opterr, optopt;
  int errflag, cp, want_stats, want_erase, want_reset; 
  int want_status, want_mrtg, want_counter, want_match, want_all_fields;
  int which_counter, fetch_from_file;
  u_int32_t what_to_count, have_wtc;

  /* Administrativia */
  memset(&q, 0, sizeof(struct query_header));
  memset(&empty_addr, 0, sizeof(struct pkt_primitives));
  memset(count, 0, sizeof(count));
  memset(&addr, 0, sizeof(struct pkt_primitives));
  memset(password, 0, sizeof(password)); 

  strcpy(path, "/tmp/collect.pipe");
  unpacked = 0; printed = 0;
  errflag = 0; buflen = 0;
  want_stats = FALSE;
  want_erase = FALSE;
  want_status = FALSE;
  want_counter = FALSE;
  want_mrtg = FALSE;
  want_match = FALSE;
  want_all_fields = FALSE;
  want_reset = FALSE;
  which_counter = FALSE;
  fetch_from_file = FALSE;
  what_to_count = FALSE;
  have_wtc = FALSE;

  while (!errflag && ((cp = getopt(argc, argv, ARGS_PMACCT)) != -1)) {
    switch (cp) {
    case 's':
      if (CHECK_Q_TYPE(q.type)) print_ex_options_error();
      q.type |= WANT_STATS;
      want_stats = TRUE;
      break;
    case 'c':
      strlcpy(count, optarg, sizeof(count));
      count_ptr = count;
      while ((*count_ptr != '\0') && (count_index <= N_PRIMITIVES-1)) {
        count_token[count_index] = extract_token(&count_ptr, ',');
	if (!strcmp(count_token[count_index], "src_host")) what_to_count |= COUNT_SRC_HOST;
        else if (!strcmp(count_token[count_index], "dst_host")) what_to_count |= COUNT_DST_HOST;
        else if (!strcmp(count_token[count_index], "sum")) what_to_count |= COUNT_SUM_HOST;
        else if (!strcmp(count_token[count_index], "src_port")) what_to_count |= COUNT_SRC_PORT;
        else if (!strcmp(count_token[count_index], "dst_port")) what_to_count |= COUNT_DST_PORT;
        else if (!strcmp(count_token[count_index], "proto")) what_to_count |= COUNT_IP_PROTO;
        else if (!strcmp(count_token[count_index], "src_mac")) what_to_count |= COUNT_SRC_MAC;
        else if (!strcmp(count_token[count_index], "dst_mac")) what_to_count |= COUNT_DST_MAC;
        else if (!strcmp(count_token[count_index], "vlan")) what_to_count |= COUNT_VLAN;
        // else if (!strcmp(count_token[count_index], "tos")) what_to_count |= COUNT_IP_TOS;
        else if (!strcmp(count_token[count_index], "none")) what_to_count |= COUNT_NONE;
        else if (!strcmp(count_token[count_index], "src_as")) what_to_count |= COUNT_SRC_AS;
        else if (!strcmp(count_token[count_index], "dst_as")) what_to_count |= COUNT_DST_AS;
        else if (!strcmp(count_token[count_index], "src_net")) what_to_count |= COUNT_SRC_NET;
        else if (!strcmp(count_token[count_index], "dst_net")) what_to_count |= COUNT_DST_NET;
        else if (!strcmp(count_token[count_index], "sum_host")) what_to_count |= COUNT_SUM_HOST;
        else if (!strcmp(count_token[count_index], "sum_net")) what_to_count |= COUNT_SUM_NET;
        else if (!strcmp(count_token[count_index], "sum_as")) what_to_count |= COUNT_SUM_AS;
        else if (!strcmp(count_token[count_index], "sum_port")) what_to_count |= COUNT_SUM_PORT;
        else printf("WARN: ignoring unknown aggregation method: %s.\n", count_token[count_index]);
	count_index++;
      }
      q.what_to_count = what_to_count;
      break;
    case 'e':
      if (CHECK_Q_TYPE(q.type)) print_ex_options_error();
      q.type |= WANT_ERASE; 
      want_erase = TRUE;
      break;
    case 't':
      if (CHECK_Q_TYPE(q.type)) print_ex_options_error();
      q.type |= WANT_STATUS; 
      want_status = TRUE;
      break;
    case 'm': /* obsoleted */
      want_mrtg = TRUE;
    case 'N':
      if (CHECK_Q_TYPE(q.type)) print_ex_options_error();
      strlcpy(match_string, optarg, sizeof(match_string));
      match_string[LARGEBUFLEN-1] = '\0';
      q.type |= WANT_COUNTER; 
      want_counter = TRUE;
      break;
    case 'n':
      if (!strcmp(optarg, "bytes")) which_counter = FALSE;
      else if (!strcmp(optarg, "packets")) which_counter = TRUE;
      else printf("WARN: ignoring unknown counter: %s.\n", optarg);
      break;
    case 'M':
      if (CHECK_Q_TYPE(q.type)) print_ex_options_error();
      strlcpy(match_string, optarg, sizeof(match_string));
      match_string[LARGEBUFLEN-1] = '\0';
      q.type |= WANT_MATCH;
      want_match = TRUE;
      break;
    case 'p':
      strlcpy(path, optarg, sizeof(path));
      break;
    case 'P':
      strlcpy(password, optarg, sizeof(password));
      break;
    case 'a':
      want_all_fields = TRUE;
      break;
    case 'r':
      q.type |= WANT_RESET;
      want_reset = TRUE;
      break;
    default:
      printf("ERROR: parameter %c unknown! \n  Exiting...\n\n", cp);
      usage_client(argv[0]);
      exit(1);
      break;
    } 
  }

  /* some post-getopt-processing task */
  if (!q.type) {
    printf("ERROR: no -s, -e, -t or -N option specified. \n  Exiting...\n\n");
    usage_client(argv[0]);
    exit(1);
  }

  if ((want_counter || want_match) && !what_to_count) {
    printf("ERROR: '-N' or '-M' selected but '-c' has not been specified or is invalid.\n  Exiting...\n\n");
    usage_client(argv[0]);
    exit(1);
  }

  if (want_reset && !(want_counter || want_match)) {
    printf("ERROR: '-r' selected but either '-N' or '-M' has not been specified.\n  Exiting...\n\n");
    usage_client(argv[0]);
    exit(1);
  }

  if (which_counter && !want_counter) {
    printf("ERROR: '-n' has to be selected in conjunction to '-N'.\n  Exiting...\n\n");
    usage_client(argv[0]);
    exit(1);
  }

  if (want_counter || want_match) {
    char *ptr = match_string, prefix[] = "file:";

    while(isspace(*ptr)) ptr++;
    if (!strncmp(ptr, prefix, strlen(prefix))) {
      fetch_from_file = TRUE; 
      ptr += strlen(prefix);
      strlcpy(file, ptr, sizeof(file)); 
    }
  }

  memcpy(q.passwd, password, sizeof(password));
  
  if (want_counter || want_match) {
    FILE *f;
    int strnum; 
    char **strings, *tptr1, *tptr2, tmpstr[SRVBUFLEN];
    char *tmpbuf, *tmpbufptr;
    
    /* 1st step: count how many queries we will have */
    if (!fetch_from_file) {
      for (strnum = 0, tptr1 = match_string; tptr1 && (strnum < MAX_QUERIES); strnum++) {
        tptr2 = tptr1;
        tptr1 = strchr(tptr1, ':'); 
        if (tptr1) {
	  if (*tptr2 == *tptr1) strnum--; /* void string */
	  tptr1++;
        }
      } 
    }
    else {
      if ((f = fopen(file, "r")) == NULL) {
        printf("ERROR: file '%s' not found\n", file);
        exit(1);
      }      
      else {
	strnum = 0;
	while (!feof(f) && (strnum < MAX_QUERIES)) {
	  if (fgets(tmpstr, SRVBUFLEN, f)) { 
	    if (!sanitize_buf(tmpstr)) strnum++;
	  }
	}
      }
    }

    strings = malloc((strnum+1)*sizeof(char *));
    if (!strings) {
      printf("ERROR: Unable to allocate sufficient memory.\n");
      exit(1); 
    }
    memset(strings, 0, (strnum+1)*sizeof(char *));

    if (fetch_from_file) {
      tmpbuf = malloc((strnum+1)*SRVBUFLEN);
      if (!tmpbuf) {
	printf("ERROR: Unable to allocate sufficient memory.\n");
	exit(1);
      }
      memset(tmpbuf, 0, (strnum+1)*SRVBUFLEN);
    }

    /* 2nd step: tokenize the whole string */
    if (!fetch_from_file) {
      for (strnum = 0, tptr1 = match_string; tptr1 && (strnum < MAX_QUERIES); strnum++) {
        tptr2 = tptr1;
        tptr1 = strchr(tptr1, ':');
        if (tptr1) *tptr1 = '\0';
        if (strlen(tptr2)) strings[strnum] = tptr2;
        else strnum--; /* void string */
        if (tptr1) tptr1++;
      }
    }
    else {
      freopen(file, "r", f);
      strnum = 0;
      tmpbufptr = tmpbuf;
      while (!feof(f) && (strnum < MAX_QUERIES)) {
        if (fgets(tmpbufptr, SRVBUFLEN, f)) {
	  tmpbufptr[SRVBUFLEN-1] = '\0';
	  if (!sanitize_buf(tmpbufptr)) {
	    strings[strnum] = tmpbufptr;
	    strnum++;
	    tmpbufptr += SRVBUFLEN;
	  }
        }
      }
      fclose(f);
    }

    bufptr = clibuf;
    bufptr += sizeof(struct query_header);
    
    /* 4th step: build queries */
    for (q.num = 0; (q.num < strnum) && (q.num < MAX_QUERIES); q.num++) {
      match_string_ptr = strings[q.num];
      match_string_index = 0;
      while ((*match_string_ptr != '\0') && (match_string_index < count_index))  {
        match_string_token = extract_token(&match_string_ptr, ',');
        if (!strcmp(count_token[match_string_index], "src_host") ||
	    !strcmp(count_token[match_string_index], "src_net") ||
	    !strcmp(count_token[match_string_index], "sum_host") ||
	    !strcmp(count_token[match_string_index], "sum_net")) {
	  if (!inet_aton(match_string_token, &addr.src_ip)) {
	    printf("ERROR: src_host: Invalid IP address: '%s'\n", match_string_token);
	    exit(1);
	  }
        }
        else if (!strcmp(count_token[match_string_index], "dst_host") ||
		 !strcmp(count_token[match_string_index], "dst_net")) {
          if (!inet_aton(match_string_token, &addr.dst_ip)) {
            printf("ERROR: dst_host: Invalid IP address: '%s'\n", match_string_token);
            exit(1);
          }
        }
        else if (!strcmp(count_token[match_string_index], "src_mac")) {
          unsigned char *eth_ptr;

          eth_ptr = (unsigned char *)ether_aton(match_string_token);
	  if (!eth_ptr) {
	    printf("ERROR: src_mac: Invalid MAC address: '%s'\n", match_string_token);
            exit(1);
	  }
	  else memcpy(&addr.eth_shost, eth_ptr, ETH_ADDR_LEN);
        }
        else if (!strcmp(count_token[match_string_index], "dst_mac")) {
          unsigned char *eth_ptr;

          eth_ptr = (unsigned char *)ether_aton(match_string_token);
          if (!eth_ptr) {
            printf("ERROR: src_mac: Invalid MAC address: '%s'\n", match_string_token);
            exit(1);
          }
          else memcpy(&addr.eth_dhost, eth_ptr, ETH_ADDR_LEN);
        }
        else if (!strcmp(count_token[match_string_index], "vlan")) {
	  addr.vlan_id = atoi(match_string_token);
        }
        else if (!strcmp(count_token[match_string_index], "src_port") ||
		 !strcmp(count_token[match_string_index], "sum_port")) { 
          addr.src_port = atoi(match_string_token);
        }
        else if (!strcmp(count_token[match_string_index], "dst_port")) {
          addr.dst_port = atoi(match_string_token);
        }
        else if (!strcmp(count_token[match_string_index], "proto")) {
	  for (index = 0; _protocols[index].number != -1; index++) { 
	    if (!strcmp(_protocols[index].name, match_string_token)) {
	      addr.proto = _protocols[index].number;
	      break;
	    }
	  }
	  if (addr.proto <= 0) {
	    addr.proto = atoi(match_string_token);
	    if ((addr.proto <= 0) || (addr.proto > 255)) {
	      printf("ERROR: invalid protocol: '%s'\n", match_string_token);
	      exit(1);
	    }
	  }
        }
	else if (!strcmp(count_token[match_string_index], "none"));
	else if (!strcmp(count_token[match_string_index], "src_as") ||
		 !strcmp(count_token[match_string_index], "sum_as")) {
	  addr.src_ip.s_addr = atoi(match_string_token);
	  addr.src_ip.s_addr = htonl(addr.src_ip.s_addr);
	}
	else if (!strcmp(count_token[match_string_index], "dst_as")) {
	  addr.dst_ip.s_addr = atoi(match_string_token);
	  addr.dst_ip.s_addr = htonl(addr.dst_ip.s_addr);
	}
        else printf("WARN: ignoring unknown aggregation method: '%s'.\n", *count_token);
        match_string_index++;
      }

      memcpy(bufptr, &addr, sizeof(struct pkt_primitives));
      bufptr += sizeof(struct pkt_primitives);
    }
  }

  /* arranging header and size of buffer to send */
  memcpy(clibuf, &q, sizeof(struct query_header)); 
  buflen = sizeof(struct query_header)+(q.num*sizeof(struct pkt_primitives));
  buflen++;
  clibuf[buflen] = '\x4'; /* EOT */
  buflen++;

  sd = build_query_client(path);
  send(sd, clibuf, buflen, 0);
  if (want_stats || want_match) {
    unpacked = Recv(sd, &largebuf);
    if (want_all_fields) have_wtc = FALSE; 
    else have_wtc = TRUE; 
    what_to_count = ((struct query_header *)largebuf)->what_to_count;
    write_stats_header(what_to_count, have_wtc);
    elem = largebuf+sizeof(struct query_header);
    unpacked -= sizeof(struct query_header);
    while (printed < unpacked) {
      acc_elem = (struct acc *) elem;
      if (memcmp(&acc_elem, &empty_addr, sizeof(struct pkt_primitives)) != 0) {
        if (!have_wtc || (what_to_count & COUNT_ID))
	  printf("%-5d  ", acc_elem->id);

	if (!have_wtc || (what_to_count & COUNT_SRC_MAC)) {
	  src_ip = (char *) ether_ntoa(acc_elem->eth_shost);
	  printf("%-17s  ", src_ip);
	}

	if (!have_wtc || (what_to_count & COUNT_DST_MAC)) {
	  src_ip = (char *) ether_ntoa(acc_elem->eth_dhost);
	  printf("%-17s  ", src_ip);
	} 

	if (!have_wtc || (what_to_count & COUNT_VLAN)) {
          printf("%-5d  ", acc_elem->vlan_id);
        }

	if (!have_wtc || (what_to_count & (COUNT_SRC_HOST|COUNT_SUM_HOST|COUNT_SRC_NET
	    |COUNT_SUM_NET|COUNT_SRC_AS|COUNT_SUM_AS))) {
	  if (what_to_count & (COUNT_SRC_AS|COUNT_SUM_AS)) printf("%-15d  ", ntohl(acc_elem->src_ip.s_addr));
	  else {
	    src_ip = inet_ntoa(acc_elem->src_ip);
	    printf("%-15s  ", src_ip);
	  }
	}

	if (!have_wtc || (what_to_count & (COUNT_DST_HOST|COUNT_DST_NET|COUNT_DST_AS))) {
	  if (what_to_count & COUNT_DST_AS) printf("%-15d  ", ntohl(acc_elem->dst_ip.s_addr));
	  else {
	    src_ip = inet_ntoa(acc_elem->dst_ip);
	    printf("%-15s  ", src_ip);
	  }
	}

	if (!have_wtc || (what_to_count & (COUNT_SRC_PORT|COUNT_SUM_PORT)))
	  printf("%-5d     ", acc_elem->src_port);

	if (!have_wtc || (what_to_count & COUNT_DST_PORT))
	  printf("%-5d     ", acc_elem->dst_port);

	if (!have_wtc || (what_to_count & COUNT_IP_PROTO))
	  printf("%-10s  ", _protocols[acc_elem->proto].name);

	//if (!have_wtc || (what_to_count & COUNT_IP_TOS))
	//  printf("%-3d    ", acc_elem->tos); 

        printf("%-10u  ", acc_elem->packet_counter); 
        printf("%u\n", acc_elem->bytes_counter); 
        counter++;
      }
      elem += sizeof(struct acc);
      printed += sizeof(struct acc);
    }
    printf("\nFor a total of: %d entries\n", counter);
  }
  else if (want_erase) printf("OK: Clearing stats.\n");
  else if (want_status) {
    unpacked = Recv(sd, &largebuf);
    write_status_header();
    elem = largebuf+sizeof(struct query_header);
    unpacked -= sizeof(struct query_header);
    while (printed < unpacked) {
      bd = (struct bucket_desc *) elem;
      printf("%u\t", bd->num);
      while (bd->howmany > 0) {
        printf("*");
	bd->howmany--;
      }
      printf("\n");

      elem += sizeof(struct bucket_desc);
      printed += sizeof(struct bucket_desc);
    }
  }
  else if (want_counter) {
    char *base, *ptr;
    int i;

    unpacked = Recv(sd, &largebuf);
    base = largebuf+sizeof(struct query_header);
    acc_elem = (struct acc *) base;
    for (i = 0; i < ((struct query_header *)largebuf)->num; i++, acc_elem++) {
      if (memcmp(&acc_elem, &empty_addr, sizeof(struct pkt_primitives)) != 0) {
        if (want_mrtg) printf("0\n"); /* input bytes: obsolete */
        if (!which_counter) printf("%d\n", acc_elem->bytes_counter); /* output bytes */
        else printf("%d\n", acc_elem->packet_counter); /* output packets */
        if (want_mrtg) {
	  printf("0\n"); /* uptime: obsolete */
          printf("0\n"); /* description: obsolete */
        }
      }
    }
  }
  else {
    usage_client(argv[0]);
    exit(1);
  }

  close(sd);

  return 0;
}

char *extract_token(char **string, int delim)
{
  char *token, *delim_ptr;

  if ((delim_ptr = strchr(*string, delim))) {
    *delim_ptr = '\0';
    token = *string;
    *string = delim_ptr+1;
  }
  else {
    token = *string;
    *string += strlen(*string);
  }

  return token;
}

int Recv(int sd, unsigned char **buf) 
{
  int num, unpacked = 0, round = 0; 
  unsigned char rxbuf[LARGEBUFLEN], *elem;

  *buf = (unsigned char *) malloc(LARGEBUFLEN);
  memset(*buf, 0, LARGEBUFLEN);
  memset(rxbuf, 0, LARGEBUFLEN);

  do {
    num = recv(sd, rxbuf, LARGEBUFLEN, 0);
    if (num > 0) {
      /* check 1: enough space in allocated buffer */
      if (unpacked+num >= round*LARGEBUFLEN) {
        round++;
        *buf = realloc((unsigned char *) *buf, round*LARGEBUFLEN);
        if (!(*buf)) {
          printf("ERROR: realloc() out of memory\n");
          exit(1);
        }
        /* ensuring realloc() didn't move somewhere else our memory area */
        elem = *buf;
        elem += unpacked;
      }
      /* check 2: enough space in dss */
      if (((u_int32_t)elem+num) > (u_int32_t)sbrk(0)) sbrk(LARGEBUFLEN);

      memcpy(elem, rxbuf, num);
      unpacked += num;
      elem += num;
    }
  } while (num > 0);

  return unpacked;
}
